// util.cpp: implementation of the Utility classes.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "util.h"
#include "snmp.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif



extern CTreeCtrl idcTree;
extern CStatic idcOid, idcName, idcType, idcValue;



PSEI sei = NULL; // ptr to SnmpExtensionInit()
PSEIE seie = NULL; // ptr to SnmpExtensionInitEx()
PSEQ seq = NULL; // ptr to SnmpExtensionQuery()



void ods( ccp const fmt, ... )
{
	char buf[2000];
	va_list args;

	va_start( args, fmt );
	_vsnprintf( buf, sizeof buf - 1, fmt, args );
	va_end( args );

	OutputDebugString( buf );
}



// AsnAny wrapper



AA::clear()
{
	asnType = ASN_NULL;
	memset( &asnValue, '\0', sizeof asnValue );
}



AA::free()
{
	switch ( asnType )
	{
	// these are all AsnOctetStrings
	case ASN_OCTETSTRING:
	case ASN_BITS:
	case ASN_SEQUENCE:
	case ASN_IPADDRESS:
	case ASN_OPAQUE:
		if ( asnValue.bits.dynamic && asnValue.bits.stream != NULL )
			SnmpUtilMemFree( asnValue.bits.stream );
		break;
	// this one is an AsnObjectIdentifier
	case ASN_OBJECTIDENTIFIER:
		if ( asnValue.object.ids != NULL )
			SnmpUtilMemFree( asnValue.object.ids );
		break;
	// all others do not have any pointers
	default:
		break;
	}
	clear();
}



AA::AA()
{
	clear();
}



AA::AA( const AA& a )
{
	UINT len;

	if ( this == &a )
		return;

	clear();

	asnType = a.asnType;

	switch ( asnType )
	{
	// these are all AsnOctetStrings
	case ASN_OCTETSTRING:
	case ASN_BITS:
	case ASN_SEQUENCE:
	case ASN_IPADDRESS:
	case ASN_OPAQUE:
		asnValue.bits.length = a.asnValue.bits.length;
		asnValue.bits.dynamic = TRUE;
		asnValue.bits.stream = (BYTE *) SnmpUtilMemAlloc( asnValue.bits.length );
		memcpy( asnValue.bits.stream, a.asnValue.bits.stream, asnValue.bits.length );
		break;
	// this one is an AsnObjectIdentifier
	case ASN_OBJECTIDENTIFIER:
		len = a.asnValue.object.idLength * sizeof a.asnValue.object.ids[0];
		asnValue.object.idLength = a.asnValue.object.idLength;
		asnValue.object.ids = (UINT *) SnmpUtilMemAlloc( len );
		memcpy( asnValue.object.ids, a.asnValue.object.ids, len );
		break;
	// all others do not have any pointers
	default:
		asnValue = a.asnValue;
		break;
	}
}



AA::AA( const AsnAny& a )
{
	UINT len;

	if ( this == &a )
		return;

	clear();

	asnType = a.asnType;

	switch ( asnType )
	{
	// these are all AsnOctetStrings
	case ASN_OCTETSTRING:
	case ASN_BITS:
	case ASN_SEQUENCE:
	case ASN_IPADDRESS:
	case ASN_OPAQUE:
		asnValue.bits.length = a.asnValue.bits.length;
		asnValue.bits.dynamic = TRUE;
		asnValue.bits.stream = (BYTE *) SnmpUtilMemAlloc( asnValue.bits.length );
		memcpy( asnValue.bits.stream, a.asnValue.bits.stream, asnValue.bits.length );
		break;
	// this one is an AsnObjectIdentifier
	case ASN_OBJECTIDENTIFIER:
		len = a.asnValue.object.idLength * sizeof a.asnValue.object.ids[0];
		asnValue.object.idLength = a.asnValue.object.idLength;
		asnValue.object.ids = (UINT *) SnmpUtilMemAlloc( len );
		memcpy( asnValue.object.ids, a.asnValue.object.ids, len );
		break;
	// all others do not have any pointers
	default:
		asnValue = a.asnValue;
		break;
	}
}



AA& AA::operator=( const AA& a )
{
	UINT len;

	if ( this == &a )
		return *this;

	free();

	asnType = a.asnType;

	switch ( asnType )
	{
	// these are all AsnOctetStrings
	case ASN_OCTETSTRING:
	case ASN_BITS:
	case ASN_SEQUENCE:
	case ASN_IPADDRESS:
	case ASN_OPAQUE:
		asnValue.bits.length = a.asnValue.bits.length;
		asnValue.bits.dynamic = TRUE;
		asnValue.bits.stream = (BYTE *) SnmpUtilMemAlloc( asnValue.bits.length );
		memcpy( asnValue.bits.stream, a.asnValue.bits.stream, asnValue.bits.length );
		break;
	// this one is an AsnObjectIdentifier
	case ASN_OBJECTIDENTIFIER:
		len = a.asnValue.object.idLength * sizeof a.asnValue.object.ids[0];
		asnValue.object.idLength = a.asnValue.object.idLength;
		asnValue.object.ids = (UINT *) SnmpUtilMemAlloc( len );
		memcpy( asnValue.object.ids, a.asnValue.object.ids, len );
		break;
	// all others do not have any pointers
	default:
		asnValue = a.asnValue;
		break;
	}

	return *this;
}



AA& AA::operator=( const AsnAny& a )
{
	UINT len;

	if ( this == &a )
		return *this;

	free();

	asnType = a.asnType;

	switch ( asnType )
	{
	// these are all AsnOctetStrings
	case ASN_OCTETSTRING:
	case ASN_BITS:
	case ASN_SEQUENCE:
	case ASN_IPADDRESS:
	case ASN_OPAQUE:
		asnValue.bits.length = a.asnValue.bits.length;
		asnValue.bits.dynamic = TRUE;
		asnValue.bits.stream = (BYTE *) SnmpUtilMemAlloc( asnValue.bits.length );
		memcpy( asnValue.bits.stream, a.asnValue.bits.stream, asnValue.bits.length );
		break;
	// this one is an AsnObjectIdentifier
	case ASN_OBJECTIDENTIFIER:
		len = a.asnValue.object.idLength * sizeof a.asnValue.object.ids[0];
		asnValue.object.idLength = a.asnValue.object.idLength;
		asnValue.object.ids = (UINT *) SnmpUtilMemAlloc( len );
		memcpy( asnValue.object.ids, a.asnValue.object.ids, len );
		break;
	// all others do not have any pointers
	default:
		asnValue = a.asnValue;
		break;
	}

	return *this;
}



AA::~AA()
{
	free();
}



AA::operator AsnAny&()
{
	return dynamic_cast<AsnAny&>( *this );
}



AA::operator const AsnAny&()
{
	return dynamic_cast<const AsnAny&>( *this );
}



AA::operator CString() const
{
	CString s( "[n/a]" );
	char *p, t[40];
	UINT i, fudge;
	bool haveAscii;

	switch ( asnType )
	{
	case 0:
		s = "[-n/a-]";
		break;
	case ASN_NULL:
		s = "[-null-]";
		break;
	case ASN_INTEGER:
		sprintf( t, "%ld", asnValue.number );
		s = t;
		break;
#ifdef ASN_COUNTER64
	case ASN_COUNTER64:
		sprintf( t, "%I64u", asnValue.counter64 );
		s = t;
		break;
#endif

	// these are all 32bit unsigned
	case ASN_COUNTER32:
#ifdef ASN_GAUGE32
	case ASN_GAUGE32:
#endif
#ifdef ASN_TIMETICKS
	case ASN_TIMETICKS:
#endif
	case ASN_UNSIGNED32:
		sprintf( t, "%lu", asnValue.unsigned32 );
		s = t;
		break;
	case ASN_IPADDRESS:
		if ( asnValue.address.length == 4 ) // looks like an IP address?
		{
			p = s.GetBuffer( 4 * 4 + 2 );
			sprintf( p, "%u.%u.%u.%u", (unsigned int) asnValue.address.stream[0],
				(unsigned int) asnValue.address.stream[1],
				(unsigned int) asnValue.address.stream[2],
				(unsigned int) asnValue.address.stream[3] );
			s.ReleaseBuffer();
			break;
		}
		// else it's not an IP address we know, fall through to hex dump
	// these are all AsnOctetStrings
	case ASN_OCTETSTRING:
		haveAscii = true;
		fudge = 0;
		p = s.GetBuffer( asnValue.bits.length * 4 + 8 );
		for ( i = 0; i < asnValue.bits.length; ++ i )
		{
			p += sprintf( p, "%02x ", asnValue.bits.stream[i] );
			if ( asnValue.bits.stream[i] < (BYTE) 0x20 )
			{
				if ( asnValue.bits.stream[i] != (BYTE) '\0' ||
					i != asnValue.bits.length - 1 ) // last char and null byte?
					haveAscii = false;
				else
					fudge = 1; // how many chars _not_ to copy to the string later
			}
		}
		if ( asnValue.bits.length > 0 )
			-- p; // skip back over last blank

		if ( haveAscii && asnValue.string.stream != NULL && asnValue.string.length != 0 )
		{
			*( p ++ ) = '\r';
			*( p ++ ) = '\n';
			*( p ++ ) = '[';
			memcpy( p, asnValue.string.stream, asnValue.string.length - fudge );
			p += asnValue.string.length - fudge;
			*( p ++ ) = ']';
		}
		*p = '\0';

		s.ReleaseBuffer();
		break;
	case ASN_BITS:
	case ASN_SEQUENCE:
	case ASN_OPAQUE:
		p = s.GetBuffer( asnValue.bits.length * 3 + 2 );
		for ( i = 0; i < asnValue.bits.length; ++ i )
		{
			p += sprintf( p, "%02x ", asnValue.bits.stream[i] );
		}
		if ( asnValue.bits.length > 0 )
			-- p; // skip back over last blank
		*p = '\0';

		s.ReleaseBuffer();
		break;
	// this one is an AsnObjectIdentifier
	case ASN_OBJECTIDENTIFIER:
		s = (CString) *( (AOI *) &asnValue.object );
		break;
	// all others do not have any pointers
	default:
		s = "[unexpected type]";
		break;
	}

	return s;
}



CString AA::getType() const
{
	ccp p;

	switch ( asnType )
	{
	case 0:
		p = "[never assigned]";
		break;
	case ASN_INTEGER:
		p = "ASN_INTEGER";
		break;
	case ASN_BITS:
		p = "ASN_BITS";
		break;
	case ASN_OCTETSTRING:
		p = "ASN_OCTETSTRING";
		break;
	case ASN_NULL:
		p = "ASN_NULL";
		break;
	case ASN_OBJECTIDENTIFIER:
		p = "ASN_OBJECTIDENTIFIER";
		break;
	case ASN_SEQUENCE:
		p = "ASN_SEQUENCE";
		break;
	case ASN_IPADDRESS:
		p = "ASN_IPADDRESS";
		break;
	case ASN_COUNTER32:
		p = "ASN_COUNTER32";
		break;
#ifdef ASN_GAUGE32
	case ASN_GAUGE32:
		p = "ASN_GAUGE32";
		break;
#endif
#ifdef ASN_TIMETICKS
	case ASN_TIMETICKS:
		p = "ASN_TIMETICKS";
		break;
#endif
	case ASN_OPAQUE:
		p = "ASN_OPAQUE";
		break;
#ifdef ASN_COUNTER64
	case ASN_COUNTER64:
		p = "ASN_COUNTER64";
		break;
#endif
	case ASN_UNSIGNED32:
		p = "ASN_UNSIGNED32";
		break;
	default:
		p = "[unknown]";
		break;
	}

	return CString( p );
}



// AsnObjectIdentifier wrapper



AOI::clear()
{
	idLength = 0;
	ids = NULL;
}



AOI::free()
{
	if ( ids != NULL )
		SnmpUtilMemFree( ids );
	clear();
}



AOI::AOI()
{
	clear();
	ids = (UINT *) SnmpUtilMemAlloc( 0 );
}



AOI::AOI( const AOI& a )
{
	if ( this == &a )
		return;

	clear();

	idLength = a.idLength;
	ids = (UINT *) SnmpUtilMemAlloc( a.idLength * sizeof ids[0] );
	memcpy( ids, a.ids, a.idLength * sizeof ids[0] );
}



AOI::AOI( const AsnObjectIdentifier& a )
{
	if ( this == &a )
		return;

	clear();

	idLength = a.idLength;
	ids = (UINT *) SnmpUtilMemAlloc( a.idLength * sizeof ids[0] );
	memcpy( ids, a.ids, a.idLength * sizeof ids[0] );
}



AOI::AOI( ccp src )
{
	char *s, *t;
	UINT i;

	clear();
	if ( src == NULL || *src == '\0' )
		return;

	for ( ; *src == '.'; ++ src )
		; // skip leading dots
	s = _strdup( src );

	for ( t = s + strlen( s ) - 1; t >= s && *t == '.'; -- t )
		; // skip trailing dots -- t points to non-dot or s-1
	*( t + 1 ) = '\0'; // truncate after last valid digit

	if ( t - s == 0 )
		return;

	for ( t = s; *t != '\0'; ++ t )
	{
		if ( *t == '.' )
			++ idLength;
	}
	++ idLength; // fencepost count + 1

	ids = (UINT *) SnmpUtilMemAlloc( idLength * sizeof ids[0] );
	for ( t = s, i = 0; i < idLength; ++ i )
	{
		ids[i] = atoi( t );
		while ( *t != '\0' && *t != '.' )
			++ t; // skip this number
		if ( *t == '.' )
			++ t; // and its trailing dot if present
	}
}



AOI::AOI( UINT n, const UINT *newids )
{
	if ( n == 0 || newids == NULL )
	{
		idLength = 0;
		ids = (UINT *) SnmpUtilMemAlloc( 0 );
		return;
	}

	idLength = n;
	ids = (UINT *) SnmpUtilMemAlloc( n * sizeof ids[0] );
	memcpy( &ids[0], &newids[0], n * sizeof ids[0] );
}



AOI& AOI::operator=( const AOI& a )
{
	if ( this == &a )
		return *this;

	idLength = a.idLength;
	ids = (UINT *) SnmpUtilMemReAlloc( ids, a.idLength * sizeof ids[0] );
	memcpy( ids, a.ids, a.idLength * sizeof ids[0] );

	return *this;
}



AOI& AOI::operator=( const AsnObjectIdentifier& a )
{
	if ( this == &a )
		return *this;

	idLength = a.idLength;
	ids = (UINT *) SnmpUtilMemReAlloc( ids, a.idLength * sizeof ids[0] );
	memcpy( ids, a.ids, a.idLength * sizeof ids[0] );

	return *this;
}



AOI& AOI::operator=( ccp src )
{
	char *s, *t;
	UINT i;

	if ( src == NULL || *src == '\0' )
	{
		idLength = 0;
		ids = (UINT *) SnmpUtilMemReAlloc( ids, 0 );
		return *this;
	}

	for ( ; *src == '.'; ++ src )
		; // skip leading dots
	s = _strdup( src );

	for ( t = s + strlen( s ) - 1; t >= s && *t == '.'; -- t )
		; // skip trailing dots -- t points to non-dot or s-1
	*( t + 1 ) = '\0'; // truncate after last valid digit

	if ( t - s == 0 )
	{
		idLength = 0;
		ids = (UINT *) SnmpUtilMemReAlloc( ids, 0 );
		return *this;
	}

	for ( t = s; *t != '\0'; ++ t )
	{
		if ( *t == '.' )
			++ idLength;
	}
	++ idLength; // fencepost count + 1

	ids = (UINT *) SnmpUtilMemReAlloc( ids, idLength * sizeof ids[0] );
	for ( t = s, i = 0; i < idLength; ++ i )
	{
		ids[i] = atoi( t );
		while ( *t != '\0' && *t != '.' )
			++ t; // skip this number
		if ( *t == '.' )
			++ t; // and its trailing dot if present
	}

	return *this;
}



AOI::~AOI()
{
	free();
}



AOI::operator AsnObjectIdentifier *()
{
	return dynamic_cast<AsnObjectIdentifier *>( this );
}



AOI::operator AsnObjectIdentifier&()
{
	return dynamic_cast<AsnObjectIdentifier&>( *this );
}



AOI::operator const AsnObjectIdentifier&() const
{
	return dynamic_cast<const AsnObjectIdentifier&>( *this );
}



AOI::operator CString() const
{
	UINT i;
	CString s;
	char *t, *p;

	s.Empty();
	p = t = (char *) malloc( 9 * idLength + 2 );
	for ( i = 0; i < idLength; ++ i )
	{
		p += sprintf( p, ".%lu", ids[i] );
	}

	s = t;
	::free( t );

	return s;
}



AOI::operator UINT() const
{
	if ( idLength == 0 )
		return 0;

	return ids[idLength - 1];
}



// Node



Node::~Node()
{
	killChildren();
}



Node::operator CString() const
{
	return (CString) aoi;
}


Node *Node::findChild( UINT what )
{
	bool found;
	int i;

	for ( found = false, i = 0; i < numChildren; ++ i )
	{
		if ( what == (UINT) children[i]->aoi )
		{
			found = true;
			break;
		}
	}

	return found? children[i]: NULL;
}



void Node::killChildren()
{
	int i;

	for ( i = 0; i < numChildren; ++ i )
		delete children[i];
	delete [] children;

	numChildren = maxChildren = 0;
	children = NULL;
}



Node *Node::addChild( const AOI& a, bool isFakeChild /* = false */ )
{
	const int CHUNKSIZE = 32;
	int i;
	Node *p;
	Node **newChildren;

	// first, check if we already have that id
	p = findChild( (UINT) a );

	if ( p == NULL ) // not there, create/add new node and copy
	{
		if ( numChildren >= maxChildren )
		{
			// need to reallocate
			newChildren = new Node *[maxChildren + CHUNKSIZE];
			memcpy( newChildren, children, numChildren * sizeof children[0] );
			memset( newChildren + numChildren, '\0', ( maxChildren - numChildren ) * sizeof newChildren[0] );
			delete [] children;
			maxChildren += CHUNKSIZE;
			children = newChildren;
		}

		i = numChildren; // remember node to write to
		++ numChildren;
		p = children[i] = new Node;
	}

	p->aoi = a;
	// set new bold/unbold state if new item, or isFake has changed
	p->isFake = isFakeChild;
	if ( p->hti == NULL )
		p->hti = idcTree.InsertItem( (ccp) (CString) *p, hti, TVI_LAST );
	idcTree.SetItemState( p->hti, p->isFake? 0: TVIS_BOLD, TVIS_BOLD );
	idcTree.SetItemData( p->hti, (DWORD) p );

	return p;
}



Node *Node::insert( AOI& a )
{
	Node *p, *cur;
	UINT i, l;
	AOI fake; // fake entry to insert intermediate nodes

	// this is _always_ called for the root node of a subtree
//	ods( "Node::insert( %s ): I am %s\n", (ccp) (CString) a, (ccp) (CString) aoi );

	// first, verify that the root node OID is a prefix of the new OID
	if ( a.idLength < aoi.idLength )
		return NULL;
	if ( SnmpUtilOidNCmp( &a, &aoi, aoi.idLength ) != 0 )
		return NULL; // not in our subtree

	// have we met the enemy, and he is we?
	if ( a.idLength == aoi.idLength )
		return this; // prefix matches, no leftovers -- must be us

	// from now on, i always indexes to the next ID to compare
	// l is the number of IDs left to check, including [i] but exluding the last one
	// p points to the next node down the tree

	for ( p = cur = this, i = aoi.idLength, l = a.idLength - aoi.idLength - 1;
		l > 0; ++ i, -- l, cur = p )
	{
//		ods( "Node::insert( %s ): looking at %s for %lu\n",
//			(ccp) (CString) a, (ccp) (CString) *cur, a.ids[i] );
		p = cur->findChild( a.ids[i] );
		if ( p == NULL ) // item doesn't exist, fake it
		{
			// set up fake so it has i+1 IDs from a
			fake = AOI( i + 1, &a.ids[0] );
//			ods( "Node::insert( %s ): faking %s\n", 
//				(ccp) (CString) a, (ccp) (CString) fake );
			// add the faked entry
			p = cur->addChild( fake, true );
		}
		else
		{
//			ods( "Node::insert( %s ): found %s\n", 
//				(ccp) (CString) a, (ccp) (CString) *p );
		}
		// we now have a matching entry, drill down or break out if far enough
	}

	// here, p points exactly at the entry of which a is (or will be) a child.

	return p->addChild( a );
}



void Node::enumSubtree()
{
	SnmpVarBindList *pvbl;
	AOI *paoi = new AOI;
	AA *paa = new AA;
	long errStatus, errIndex;
	Node *n;
#ifdef NO_TRAILING_ZERO_IN_OID
	bool haveTrailingZero;
#endif

	pvbl = (SnmpVarBindList *) SnmpUtilMemAlloc( sizeof SnmpVarBindList );
	pvbl->len = 1;
	pvbl->list = (SnmpVarBind *) SnmpUtilMemAlloc( sizeof SnmpVarBind );
	memset( pvbl->list, '\0', sizeof SnmpVarBind );

	// why do I get away with the following? Because none of the classes
	// have a vtable or any data members besides those in their base classes
	// (AsnObjectIdentifier, AsnAny). And yes, it's an _ugly_ hack.
	paoi = (AOI *) &pvbl->list->name;
	paa = (AA *) &pvbl->list->value;

	// do a GET for the current value
	*paoi = aoi;
	if ( seq( ASN_RFC1157_GETREQUEST, pvbl, &errStatus, &errIndex ) )
	{
		n = insert( *paoi );
		if ( n == NULL )
			ods( "Node::enumSubtree(): insert( %s ) returned NULL\n",
				(ccp) (CString) *paoi );
		else
			n->val = *paa; // save a copy of the value
	}
	else
	{
		ods( "SEQ( GET, %s ): status = %ld, index = %ld\n",
			(ccp) (CString) *paoi, errStatus, errIndex );
	}

	*paoi = aoi;
	while ( 1 )
	{
		// do a GETNEXT for the next OID
		if ( seq( ASN_RFC1157_GETNEXTREQUEST, pvbl, &errStatus, &errIndex ) )
		{
			if ( errStatus == SNMP_ERRORSTATUS_NOSUCHNAME ||
				SnmpUtilOidNCmp( paoi, aoi, aoi.idLength ) != 0 )
				break;

#ifdef NO_TRAILING_ZERO_IN_OID
			haveTrailingZero = ( paoi->ids[paoi->idLength - 1] == 0 );
			if ( haveTrailingZero )
			{
				ods( "truncating trailing zero in AOI: %s\n", (ccp) (CString) *paoi );
				-- paoi->idLength;
			}
#endif

			n = insert( *paoi );
			if ( n == NULL )
				ods( "Node::enumSubtree(): insert( %s ) returned NULL\n",
					(ccp) (CString) *paoi );
			else
				n->val = *paa; // save a copy of the value

#ifdef NO_TRAILING_ZERO_IN_OID
			if ( haveTrailingZero )
			{
				ods( "restoring trailing zero in AOI: %s\n", (ccp) (CString) *paoi );
				++ paoi->idLength;
			}
#endif
		}
		else
		{
			ods( "SEQ( GETNEXT, %s ): status = %ld, index = %ld\n",
				(ccp) (CString) *paoi, errStatus, errIndex );
			break;
		}
	}

	paoi->free();
	paa->free();
	SnmpUtilMemFree( pvbl->list );
	SnmpUtilMemFree( pvbl );
}
